<?php
namespace nur\sery\db\sqlite;

use Generator;
use nur\sery\cl;
use SQLite3;
use SQLite3Result;
use SQLite3Stmt;

/**
 * Class Sqlite: frontend vers une base de données sqlite3
 */
class Sqlite {
  static function with($sqlite, ?array $params=null): self {
    if ($sqlite instanceof static) {
      return $sqlite;
    } elseif ($sqlite instanceof self) {
      # recréer avec les mêmes paramètres
      return new static(null, cl::merge([
        "file" => $sqlite->file,
        "flags" => $sqlite->flags,
        "encryption_key" => $sqlite->encryptionKey,
        "allow_wal" => $sqlite->allowWal,
        "config" => $sqlite->config,
        "migrate" => $sqlite->migration,
      ], $params));
    } elseif (is_array($sqlite)) {
      return new static(null, cl::merge($sqlite, $params));
    } else {
      return new static($sqlite, $params);
    }
  }

  static function config_enableExceptions(self $sqlite): void {
    $sqlite->db->enableExceptions(true);
  }

  static function config_enableWalIfAllowed(self $sqlite): void {
    if ($sqlite->isWalAllowed()) {
      $sqlite->db->exec("PRAGMA journal_mode=WAL");
    }
  }

  const ALLOW_WAL = null;

  const CONFIG = [
    [self::class, "config_enableExceptions"],
    [self::class, "config_enableWalIfAllowed"],
  ];

  const MIGRATE = null;

  const params_SCHEMA = [
    "file" => ["string", ""],
    "flags" => ["int", SQLITE3_OPEN_READWRITE + SQLITE3_OPEN_CREATE],
    "encryption_key" => ["string", ""],
    "allow_wal" => ["?bool"],
    "config" => ["?array|callable"],
    "migrate" => ["?array|string|callable"],
    "auto_open" => ["bool", true],
  ];

  function __construct(?string $file=null, ?array $params=null) {
    if ($file !== null) $params["file"] = $file;
    ##schéma
    $defaultFile = self::params_SCHEMA["file"][1];
    $this->file = $file = strval($params["file"] ?? $defaultFile);
    $inMemory = $file === ":memory:";
    #
    $defaultFlags = self::params_SCHEMA["flags"][1];
    $this->flags = intval($params["flags"] ?? $defaultFlags);
    #
    $defaultEncryptionKey = self::params_SCHEMA["encryption_key"][1];
    $this->encryptionKey = strval($params["encryption_key"] ?? $defaultEncryptionKey);
    #
    $defaultAllowWal = static::ALLOW_WAL ?? !$inMemory;
    $this->allowWal = $params["allow_wal"] ?? $defaultAllowWal;
    # configuration
    $this->config = $params["config"] ?? static::CONFIG;
    # migrations
    $this->migration = $params["migrate"] ?? static::MIGRATE;
    #
    $defaultAutoOpen = self::params_SCHEMA["auto_open"][1];
    if ($params["auto_open"] ?? $defaultAutoOpen) {
      $this->open();
    }
  }

  /** @var string */
  protected $file;

  /** @var int */
  protected $flags;

  /** @var string */
  protected $encryptionKey;

  /** @var bool */
  protected $allowWal;

  /** vérifier s'il est autorisé de configurer le mode WAL */
  function isWalAllowed(): bool {
    return $this->allowWal;
  }

  /** @var array|string|callable */
  protected $config;

  /** @var array|string|callable */
  protected $migration;

  /** @var SQLite3 */
  protected $db;

  function open(): self {
    if ($this->db === null) {
      $this->db = new SQLite3($this->file, $this->flags, $this->encryptionKey);
      _config::with($this->config)->configure($this);
      _migration::with($this->migration)->migrate($this);
    }
    return $this;
  }

  function close(): void {
    if ($this->db !== null) {
      $this->db->close();
      $this->db = null;
    }
  }

  protected function checkStmt($stmt): SQLite3Stmt {
    return SqliteException::check($this->db, $stmt);
  }

  protected function checkResult($result): SQLite3Result {
    return SqliteException::check($this->db, $result);
  }

  protected function db(): SQLite3 {
    $this->open();
    return $this->db;
  }

  function _exec(string $query): bool {
    return $this->db()->exec($query);
  }

  function exec($query, ?array $params=null): bool {
    $db = $this->db();
    $query = new _query_base($query, $params);
    if ($query->useStmt($db, $stmt, $sql)) {
      try {
        return $stmt->execute()->finalize();
      } finally {
        $stmt->close();
      }
    } else {
      return $db->exec($sql);
    }
  }

  function beginTransaction(): void {
    $this->db()->exec("begin");
  }

  function commit(): void {
    $this->db()->exec("commit");
  }

  function rollback(): void {
    $this->db()->exec("commit");
  }

  function _get(string $query, bool $entireRow=false) {
    return $this->db()->querySingle($query, $entireRow);
  }

  function get($query, ?array $params=null, bool $entireRow=false) {
    $db = $this->db();
    $query = new _query_base($query, $params);
    if ($query->useStmt($db, $stmt, $sql)) {
      try {
        $result = $this->checkResult($stmt->execute());
        try {
          $row = $result->fetchArray(SQLITE3_ASSOC);
          if ($row === false) return null;
          elseif ($entireRow) return $row;
          else return cl::first($row);
        } finally {
          $result->finalize();
        }
      } finally {
        $stmt->close();
      }
    } else {
      return $db->querySingle($sql, $entireRow);
    }
  }

  function one($query, ?array $params=null): ?array {
    return $this->get($query, $params, true);
  }

  protected function _fetchResult(SQLite3Result $result, ?SQLite3Stmt $stmt=null, $primaryKeys=null): Generator {
    if ($primaryKeys !== null) $primaryKeys = cl::with($primaryKeys);
    try {
      while (($row = $result->fetchArray(SQLITE3_ASSOC)) !== false) {
        if ($primaryKeys !== null) {
          $key = implode("-", cl::select($row, $primaryKeys));
          yield $key => $row;
        } else {
          yield $row;
        }
      }
    } finally {
      $result->finalize();
      if ($stmt !== null) $stmt->close();
    }
  }

  /**
   * si $primaryKeys est fourni, le résultat est indexé sur la(es) colonne(s)
   * spécifiée(s)
   */
  function all($query, ?array $params=null, $primaryKeys=null): iterable {
    $db = $this->db();
    $query = new _query_base($query, $params);
    if ($query->useStmt($db, $stmt, $sql)) {
      $result = $this->checkResult($stmt->execute());
      return $this->_fetchResult($result, $stmt, $primaryKeys);
    } else {
      $result = $this->checkResult($db->query($sql));
      return $this->_fetchResult($result, null, $primaryKeys);
    }
  }
}
